/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fmllist.c,v $
 * Revision 1.5  1995/05/16 09:39:33  adam
 * LICENSE.
 *
 * Revision 1.4  1995/02/23  08:32:05  adam
 * Changed header.
 *
 * Revision 1.2  1995/02/10  15:50:55  adam
 * MARC interface implemented. Minor bugs fixed. fmltest can
 * be used to format single MARC records. New function '\list'
 * implemented.
 *
 * Revision 1.1  1995/02/09  14:33:37  adam
 * Split source fml.c and define relevant build-in functions in separate
 * files. New operators mult, div, not, llen implemented.
 *
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "fmlp.h"

static struct fml_node *fml_exec_indx (Fml fml, struct fml_node *l,
                                       struct fml_node *r)
{
    struct fml_node *list = l;
    struct fml_node *fn;
    int indx;

    if (!l || !r || !r->is_atom)
    {
        fml_node_delete (fml, l);
        fml_node_delete (fml, r);
        return NULL;
    }
    indx = fml_atom_val (r->p[0]);
    fml_node_delete (fml, r);
    while (--indx >= 1 && list)
        list = list->p[1];
    if (!list)
        fn = NULL;
    else if (list->is_atom)
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = list->p[0];
        list->is_atom = 0;
        list->p[0] = NULL;
    }
    else
    {
        fn = list->p[0];
        list->p[0] = NULL;
    }
    fml_node_delete (fml, l);
    return fn;
}

static struct fml_node *fml_exec_len (Fml fml, struct fml_node **lp, 
                                      struct token *tp)
{
    struct fml_node *fn;
    int len = 0;
    char arg[32];

    fml_cmd_lex (lp, tp);
    if (tp->kind == 'g')
    {
        for (fn = tp->sub; fn; fn = fn->p[1])
            len++;
    }
    else if (tp->kind == 'e')
    {
        struct fml_sym_info *info;

        info = fml_sym_lookup (fml->sym_tab, tp->tokenbuf);
        assert (info);
        for (fn = info->body; fn; fn = fn->p[1])
            len++;
    }
    else 
        len = 1;

    sprintf (arg, "%d", len);
    fn = fml_node_alloc (fml);
    fn->is_atom = 1;
    fn->p[0] = fml_atom_alloc (fml, arg);
    fml_cmd_lex (lp, tp);
    return fn;
}

static struct fml_node *fml_exec_list (Fml fml, struct fml_node **lp,
                                       struct token *tp)
{
    struct fml_node *fn = NULL;

    fml_cmd_lex (lp, tp);
    if (tp->kind == 'g')
        fn = fml_node_copy (fml, tp->sub);
    else
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, tp->tokenbuf);
    }
    fml_cmd_lex (lp, tp);
    return fn;
}

void fml_list_init (Fml fml)
{
    struct fml_sym_info *sym_info;

    sym_info = fml_sym_add (fml->sym_tab, "index");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_indx;

    sym_info = fml_sym_add (fml->sym_tab, "llen");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_len;

    sym_info = fml_sym_add (fml->sym_tab, "list");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_list;
}
