/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fmlmarc.c,v $
 * Revision 1.6  1995/05/16 09:39:33  adam
 * LICENSE.
 *
 * Revision 1.5  1995/03/30  14:22:02  adam
 * Uses MARC anchor functions now.
 *
 * Revision 1.4  1995/02/23  08:32:05  adam
 * Changed header.
 *
 * Revision 1.2  1995/02/10  16:52:08  adam
 * Indicator field moved in MARC structure. The FML list representation
 * of a MARC record has changed.
 *
 * Revision 1.1  1995/02/10  15:50:56  adam
 * MARC interface implemented. Minor bugs fixed. fmltest can
 * be used to format single MARC records. New function '\list'
 * implemented.
 *
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include <fmlmarc.h>
#include <iso2709.h>

#include "fmlp.h"

static void add_string (const char *str, char **buf, int *max, int *size)
{
    if (*size + strlen(str) >= *max)
    {
        char *nbuf;
        int nsize = *size + strlen(str) + 2048;

        nbuf = malloc (nsize);
        assert (nbuf);
        if (*buf)
            strcpy (nbuf, *buf);
        else
            *nbuf = '\0';
        free (*buf);
        *buf = nbuf;
        *max = nsize;
    }
    strcpy (*buf + *size, str);
    *size += strlen(str);
}

char *marc_to_str (Fml fml, Iso2709Rec rec)
{
    static char *buf = NULL;
    static int max = 0;
    int size = 0;
    Iso2709Anchor a;
    char *tag, *indicator, *identifier, *data;

    add_string ("{", &buf, &max, &size);
    a = iso2709_a_mk (rec);
    do
    {
        if (!iso2709_a_info_line (a, &tag, &indicator))
            break;
        add_string ("{\'", &buf, &max, &size);
        add_string (tag, &buf, &max, &size);
        add_string ("\'",&buf, &max, &size);
        if (indicator)
        {
            add_string ("\'", &buf, &max, &size);
            add_string (indicator, &buf, &max, &size);
            add_string ("\'", &buf, &max, &size);
        }
        else
            add_string ("{}", &buf, &max, &size);
        add_string ("{", &buf, &max, &size);
        do
        {
            iso2709_a_info_field (a, NULL, NULL, &identifier, &data);
            add_string ("{", &buf, &max, &size);
            if (identifier)
            {
                add_string ("\'", &buf, &max, &size);
                add_string (identifier, &buf, &max, &size);
                add_string ("\'", &buf, &max, &size);
            }
            else
                add_string ("{}", &buf, &max, &size);
            add_string (" \'", &buf, &max, &size);
            add_string (data, &buf, &max, &size);
            add_string ("\'}", &buf, &max, &size);
        } while (iso2709_a_next_field (a));
        add_string ("}}\n", &buf, &max, &size);
    } while (iso2709_a_next_line (a));
    add_string ("}", &buf, &max, &size);
    iso2709_a_rm (a);
    return buf;
}
