/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fmlrel.c,v $
 * Revision 1.4  1995/05/16 09:39:34  adam
 * LICENSE.
 *
 * Revision 1.3  1995/02/23  08:32:06  adam
 * Changed header.
 *
 * Revision 1.1  1995/02/09  14:33:37  adam
 * Split source fml.c and define relevant build-in functions in separate
 * files. New operators mult, div, not, llen implemented.
 *
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "fmlp.h"

static struct fml_node *fml_exec_gt (Fml fml, struct fml_node *l,
                                     struct fml_node *r)
{
    int left_val, right_val;
    struct fml_node *fn;
    fml_lr_values (fml, l, &left_val, r, &right_val);
    if (left_val > right_val)
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, "1");
    }
    else
        fn = NULL;
    return fn;
}

static struct fml_node *fml_exec_lt (Fml fml, struct fml_node *l,
                                     struct fml_node *r)
{
    int left_val, right_val;
    struct fml_node *fn;
    fml_lr_values (fml, l, &left_val, r, &right_val);
    if (left_val < right_val)
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, "1");
    }
    else
        fn = NULL;
    return fn;
}

static struct fml_node *fml_exec_eq (Fml fml, struct fml_node *l,
                                     struct fml_node *r)
{
    int left_val, right_val;
    struct fml_node *fn;
    fml_lr_values (fml, l, &left_val, r, &right_val);
    if (left_val == right_val)
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, "1");
    }
    else
        fn = NULL;
    return fn;
}

static struct fml_node *fml_exec_ne (Fml fml, struct fml_node *l,
                                     struct fml_node *r)
{
    int left_val, right_val;
    struct fml_node *fn;
    fml_lr_values (fml, l, &left_val, r, &right_val);
    if (left_val != right_val)
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, "1");
    }
    else
        fn = NULL;
    return fn;
}

static struct fml_node *fml_exec_le (Fml fml, struct fml_node *l,
                                     struct fml_node *r)
{
    int left_val, right_val;
    struct fml_node *fn;
    fml_lr_values (fml, l, &left_val, r, &right_val);
    if (left_val <= right_val)
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, "1");
    }
    else
        fn = NULL;
    return fn;
}

static struct fml_node *fml_exec_ge (Fml fml, struct fml_node *l,
                                     struct fml_node *r)
{
    int left_val, right_val;
    struct fml_node *fn;
    fml_lr_values (fml, l, &left_val, r, &right_val);
    if (left_val >= right_val)
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, "1");
    }
    else
        fn = NULL;
    return fn;
}

static struct fml_node *fml_exec_and (Fml fml, struct fml_node *l,
                                      struct fml_node *r)
{
    if (l && r)
    {
        fml_node_delete (fml, l);
        return r;
    }
    fml_node_delete (fml, l);
    fml_node_delete (fml, r);
    return NULL;
}

static struct fml_node *fml_exec_or (Fml fml, struct fml_node *l,
                                      struct fml_node *r)
{
    if (r)
    {
        fml_node_delete (fml, l);
        return r;
    }
    return l;
}

static struct fml_node *fml_exec_not (Fml fml, struct fml_node **lp, 
                                      struct token *tp)
{
    struct fml_node *fn;
    fml_cmd_lex (lp, tp);

    fn = fml_expr_term (fml, lp, tp);
    if (fn)
    {
        fml_node_delete (fml, fn);
        return NULL;
    }
    fn = fml_node_alloc (fml);
    fn->is_atom = 1;
    fn->p[0] = fml_atom_alloc (fml, "1");
    return fn;
}

void fml_rel_init (Fml fml)
{
    struct fml_sym_info *sym_info;

    sym_info = fml_sym_add (fml->sym_tab, "gt");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_gt;
    sym_info = fml_sym_add (fml->sym_tab, "lt");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_lt;
    sym_info = fml_sym_add (fml->sym_tab, "eq");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_eq;

    sym_info = fml_sym_add (fml->sym_tab, "ge");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_ge;
    sym_info = fml_sym_add (fml->sym_tab, "le");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_le;
    sym_info = fml_sym_add (fml->sym_tab, "ne");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_ne;

    sym_info = fml_sym_add (fml->sym_tab, "and");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_and;
    sym_info = fml_sym_add (fml->sym_tab, "or");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_or;

    sym_info = fml_sym_add (fml->sym_tab, "not");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_not;
}
