/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*
 * Resource management.
 *
 * Europagate, 1994-1995.
 *
 * $Log: gw-res.h,v $
 * Revision 1.6  1995/05/16 09:39:39  adam
 * LICENSE.
 *
 * Revision 1.5  1995/05/03  07:36:29  adam
 * New functions: gw_res_bool and gw_res_int.
 *
 * Revision 1.4  1995/04/19  12:12:02  adam
 * Resource system uses only one log debug level.
 *
 * Revision 1.3  1995/02/23  08:32:12  adam
 * Changed header.
 *
 * Revision 1.1.1.1  1995/02/09  17:27:12  adam
 * Initial version of email gateway under CVS control.
 *
 * Initial:       Dec  7, 94 (Adam Dickmeiss)
 */
#ifndef GW_RES_H
#define GW_RES_H

#define RES_DEBUG GW_LOG_DEBUGN(8)

typedef struct Gw_res_info *GwRes;   /* Gateway resource handle */

GwRes gw_res_init (void);
/*
    A resource handle is returned by this function describing
    the empty resource.
 */

void gw_res_close (GwRes id);
/*
    The resources described by 'id' are freed. No further references
    to 'id' are allowed.
 */

int gw_res_merge (GwRes id, const char *filename);
/*
    The resources described by 'id' are merged by the contents of 
    'filename'. If a resource is duplicated (in both resources 'id' and
    the file) the resource is set to the value specified in 'filename'.

    This function returns 0 on success; -1 on failure ('filename'
    could not be read)
 */

const char *gw_res_get (GwRes id, const char *name, const char *def);
/*
    The resource with name 'name' is checked in the resources represented
    by 'id'. If the resource is present a pointer to the value (null-
    terminated string) is returned. If the value is not present the 
    value of 'def' is returned.
 */

int gw_res_put (GwRes id, const char *name, const char *value, 
                const char *fname);
/*
    Change a resource - modify if it exists - add if not already
    there. The resource will have impact on the file name 'fname'.
    Use gw_res_commit (see below) to actually write to the
    resource file.
 */

int gw_res_commit (GwRes id, const char *fname);
/*
    Rewrite the resource file 'fname'. If resources are modified/added
    then these will be written now.
 */


int gw_res_trav (GwRes id, const char *fname, void (*tf)(const char *name,
                                                        const char *value));
/*
    Traverse resources associated with file 'fname'. For each resource
    the handler 'tf' is invoked with name and value.
 */

int gw_res_bool (GwRes res, const char *name, int def_val);
/*
    Return true(1) or false(0) depending on setting for name. If
    name doesn't exist, def_val is returned
 */

int gw_res_int (GwRes res, const char *name, int def_val);
/*
    Treat value of name as integer. Issue warning (GW_LOG_WARN) if the value
    cannot be read as integer. If name doesn't exist, def_val is returned
 */
/*
   GwRes file format.

   A resource name must begin on column 0 on a  line. The name is followed 
   by colon. The value of the resource comes after the colon. A value may 
   span over several lines. Subsequent value lines are preceeded by one or 
   more blanks (tab/space). Empty/blank lines are ignored. Lines beginning 
   with # are treated as comments. 

   Example
    # Comment. Ignored
   
    # Single line resource
    MaxSize: 1000
    # Multi line resource
    FatalMsg: A serious error
     occured. Aborting.
    # Yet another (danish):
    Warning: Advarsel, taenk
         dig om - inden du foretager dig noget.

   The FatalMsg resource has the value "A serious error occured. Aborting.", 
   and the Warning resource has the value 
   "Advarsel, taenk dig om - inden du foretager dig noget." Note
   that all blanks used to separate subsequent lines are treated as exactly 
   one blank.
*/
#endif
