/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * Test of resource management.
 *
 * Europagate, 1994-1995.
 *
 * $Log: gw-res-test.c,v $
 * Revision 1.4  1995/05/16 09:40:49  adam
 * LICENSE.
 *
 * Revision 1.3  1995/02/23  08:32:22  adam
 * Changed header.
 *
 * Revision 1.1.1.1  1995/02/09  17:27:12  adam
 * Initial version of email gateway under CVS control.
 *
 * Initial:       Dec  7, 94 (Adam Dickmeiss)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include <gw-log.h>
#include <gw-res.h>

static GwRes res;

static void showf (const char *name, const char *value)
{
    printf ("%s: %s\n", name, value);
}

static void interactive (void)
{
    char buffer[128];
    char arg[3][128];
    int no_arg;
    int r;

    while (1)
    {
        char *cp;
        const char *value;

        putc ('>', stdout);
        fflush (stdout);
        if (!(fgets (buffer, sizeof(buffer)-1, stdin)))
            break;
        while ((cp = strchr (buffer, '\n')))
            *cp = '\0';
        if (*buffer == '\0')
            continue;
        no_arg = sscanf (buffer+1, "%s %s %s", arg[0], arg[1], arg[2]);
        r = -10;
        switch (*buffer)
        {
        case 'm':
            if (no_arg == 1)
                r = gw_res_merge (res, arg[0]);
            break;
        case 'p':
            if (no_arg == 3)
                r = gw_res_put (res, arg[0], arg[1], arg[2]);
            break;
        case 'g':
            if (no_arg == 1)
            {
                value = gw_res_get (res, arg[0], "<unknown>");
                printf ("%s has value `%s'\n", arg[0], value);
                r = 0;
            }
            break;
        case 's':
            if (no_arg == 1)
                r = gw_res_trav (res, arg[0], showf);
            else if (no_arg <= 0)
                r = gw_res_trav (res, NULL, showf);
            break;
        case 'c':
            if (no_arg == 1)
                r = gw_res_commit (res, arg[0]);
            break;
        case 'h':
            printf ("m <file>                     merge\n"
                    "g <name>                     get\n"
                    "p <name> <value> <file>      put\n"
                    "s [<file>]                   show\n"
                    "c <file>                     commit\n"
                    "q                            quit\n");
            r = 0;
            break;
        case 'q':
            return;
        }
        if (r == -10)
            printf ("syntax, type 'h' for help\n");
        else if (r)
            printf ("returned %d\n", r);
        else
            printf ("ok\n");
    }
}

static void do_intensive (void)
{
    int i;
    for (i = 0; i<100; i++)
    {
        const char *cp;
        char new_val[12];
        res = gw_res_init ();
        
        gw_res_merge (res, "big.res");
        gw_res_merge (res, "default.res");
        gw_res_merge (res, "adam.res");

        cp = gw_res_get (res, "loopvar", "0");
        sprintf (new_val, "%d", atoi(cp)+1);
        gw_res_put (res, "loopvar", new_val, "big.res");

        gw_res_commit (res, "big.res");
        gw_res_close (res);
    }
}

int main (int argc, char **argv)
{
    gw_log_init (*argv);

    while (--argc > 0)
    {
        if (**++argv == '-')
        {
            switch (*++*argv)
            {
            case 'd':
                gw_log_level (GW_LOG_ALL);
                continue;
            case 't':
                do_intensive ();
                return 0;
            default:
                gw_log (GW_LOG_FATAL, "Fatal", "Unknown option %s",
                        *argv);
                exit (1);
            }
        }
    }
    gw_log_file (GW_LOG_ALL, "all.log");

    res = gw_res_init ();

    interactive ();

    gw_res_close (res);
    exit (0);
}
